/*
 * StartVM.java
 *
 * Created on August 31, 2000, 3:16 PM
 */

package project.manager;

import java.io.FileNotFoundException;
import java.io.File;
import java.util.*;
import java.rmi.*;
import java.rmi.activation.*;
import java.rmi.server.*;
import java.rmi.registry.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
/**
 *
 * @author  alastair
 * @version 
 */
public class RegisterStub extends Object implements java.io.Serializable {
	private final static String JAVA_SECURITY_POLICY = "java.security.policy";
	private String _security_policy = null;
	private String _code_base = null;
	private String _activation_description = null;
	private String _name_server = null;
	private String _vm_name = null;
	
	/** This class setsup and registers a remote activatable RMI server
	 * that has the ability to controle an agents behaviour eg. start, stop, suspend, resume and get its state,
	 *
	 * @param profile_location a URL where the profile can be located
	 */
    public RegisterStub(String activation_description, String naming_service, String vm_name) {
        _activation_description = activation_description;
        _name_server = naming_service;
        _vm_name = vm_name;
        _security_policy = System.getProperty( "java.security.policy" );
	    _code_base = System.getProperty( "java.rmi.server.codebase" );
        System.out.println("policy = " + _security_policy );
        System.out.println("code_base = " + _code_base );
        System.out.println("activation_description = " + _activation_description );
        System.out.println("name_server = " + _name_server );
        System.setSecurityManager( new RMISecurityManager() );
        Thread thread = Thread.currentThread();
        try
        {
            ClassLoader loader = RMIClassLoader.getClassLoader( System.getProperty( "java.rmi.server.codebase" ) );
            thread.setContextClassLoader( loader );
        }
        catch( Exception e )
        {
            e.printStackTrace();
        }
		registerVM();
    }

	private void registerVM()
	{
		System.setSecurityManager( new RMISecurityManager() );

		Properties props = new Properties(); 
		//can use the System.getProperty() mehtod but here its retrieved from a profile
		 props.put( JAVA_SECURITY_POLICY, _security_policy); 

		ActivationGroupDesc.CommandEnvironment ace = null; 
		ActivationGroupDesc activation_group = new ActivationGroupDesc(props, ace);
		VMCommsInterface vm = null;
		try
		{
		// Once the ActivationGroupDesc has been created, register it 
		// with the activation system to obtain its ID
		//
		ActivationSystem system = ActivationGroup.getSystem();
		ActivationGroupID agi = system.registerGroup( activation_group );
		
		// Now explicitly create the group
		//
		ActivationGroup.createGroup(agi, activation_group, 0);

		// Create the rest of the parameters that will be passed to
		// the ActivationDesc constructor
		//
		MarshalledObject data = null;

		// The second argument to the ActivationDesc constructor will be used 
		// to uniquely identify this class; it's location is relative to the  
		// URL-formatted String,  _code_base.
		
		ActivationDesc desc = new ActivationDesc ( _activation_description, _code_base, data);

		vm = (VMCommsInterface)Activatable.register(desc);
		System.out.println( "Got the stub for the VMRemote" + _vm_name);
		System.out.println( "Name Server = " + _name_server );
        System.out.println( _name_server + "VMRemote" + _vm_name);

		// Bind the stub to a name in the registry		
		
		Naming.rebind( _name_server + "VMRemote" + _vm_name, vm);
		System.out.println("Exported VMRemote " + _vm_name);
		}
		//todo catch the exeptions correctly and handle appropriately
		catch(Throwable t )
		{
			t.printStackTrace();
		}
		System.exit(0);
	}
	
	/**
	 * @param args args[0] requires a URL location of the profile
	 */
	public static void main( String args[] ){
		if( args.length != 3)
		{
			System.out.println("Usage:  fipaos.platform.manager.RegisterStub <class><naming service name><name to bind>" );
			
		}
        else
        {
            RegisterStub  vm = new RegisterStub( args[0], args[1], args[2] );
        }
	}
}