/*
 * VMManager.java
 *
 * Created on August 13, 2000, 9:00 AM
 */
 
package project.manager;
import fipaos.agent.FIPAOSAgent;

import fipaos.agent.profile.PlatformProfile;
import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.ont.BASETOKENS;
import fipaos.parser.validating.documentrepository.*;
import fipaos.agent.conversation.content.Content;
import fipaos.agent.conversation.content.ContentUtil;
import fipaos.parser.rdf.RDFContentHandler;
import fipaos.util.DIAGNOSTICS;
import fipaos.ont.fipa.fipaman.AgentID;


import java.lang.reflect.*;
import java.util.*;
import java.io.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.URL;
import java.net.MalformedURLException;
import java.rmi.*;
import java.rmi.activation.*;
import java.rmi.server.*;
import java.rmi.registry.*;
/**
 * @author Alastair Duncan
 * @version
 */
public class VMPlatformManager extends Object implements java.io.Serializable {
	
  
	private static String AGENTS_BAG 			= "agents";
	private static String SCHEMA 				= "http://www.nortelnetworks.com/fipa-os/schemas/Objects#";
	private static String START                 = SCHEMA + "start";
	private static String CLASS_NAME 			= SCHEMA + "class";
	private static String AGENT_NAME 			= SCHEMA + "agent_name";
	private static String PLATFORM_PROFILE_LOCATION = SCHEMA + "platform_profile";
	private static String OWNER 				= SCHEMA + "owner";
	private static String SPACE 				= " ";
	private static String PLATFORM_PROFILE		= "platform.profile";
	private static String PROFILE				= ".profile";
	private static String HOST					= SCHEMA + "host";
    private static String LISTENER_TASK					= SCHEMA + "listener_task";
	private static String TRUE					= "true";
	private static String PLATFORM_VM			= "platformVM";
	private static String VM_REMOTE				= "VMRemote";
   

//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

   // Contains the location of the platform profile file (ie. "c:\\fipa-os\\profiles")
   /** The location of the platform profile
   */
	private String _platform_location = new String();
   
   // Contains the relative name of the profile - (ie. "platform.profile").
   /** file name of the profile
 */
	private String _profile_file_name = new String();
   /** Hashtable of instances of started agents (key = name)(object = instance)
   */
	private Hashtable _agents = new Hashtable();
   /** List of agents names
   */
	private Vector _agent_list = new Vector();

  	/**
	*hash table of vms
	*/
	private Hashtable _remote_vms = new Hashtable();
	/** Content of the profile file
	*/
	private Content _doc_content = null;
   /** Vector of vectors representint a table of the agents details
   */
	private Vector _cell_data = new Vector();
   /** Vector of Threadgroups - one threadgroup for each agent
   */
	private Vector _agent_threads;
	//location of the RMI Name server
	private String _name_server_location;
	private String _name_service_port;
	private String _platform_profile;
	//the key will be the agents name the object will be a reference
	//to the vm that the agent is running in 
	private Hashtable _hosts;
	
//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------
  
  /** Constructs the VM Manager and instantiates the specified agents.
   * @param profile_location Directory where the agent loader profile is stored
   * @param file_name the filename of the profile
   */
	public VMPlatformManager( String profile_location, String file_name ) 
	{
        //this will be used by the where are you task so that it can find out if its on the platformVM or a remote vm 
        System.setProperty( "vm", "platformVM" );
		_platform_profile = profile_location;
        
		try
		{
			PlatformProfile profile = new PlatformProfile( profile_location );
			_platform_location = profile.getProfileDirectory();
		}
		catch( Throwable t )
		{
			DIAGNOSTICS.println( "Error reading platform profile - aborting", t, DIAGNOSTICS.LEVEL_MAX );
			return;
		}
		
		_agent_threads = new Vector();
		_hosts = new Hashtable();
		initCellData( _platform_location, file_name );
		startAgents();
		//VMPlatformGUI managerGUI = new VMPlatformGUI( this );
		//managerGUI.pack();
		//managerGUI.show();
  }
//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------
  
	/** Adds the details of an agent to Vector of Agents used for updating the table of agents
	*  used used by the add Agent dialog
	   * @param agent The agent to be added
	   */
	  public void addAgent(Vector agent) {
		_cell_data.add( agent );
	  }
  
	  /** Adds the agent name to the list of agents and the hashtable of agent instances
	   * @param agent_name the object representing the agent name
	   * @param instance an object array with the class name in element 0 and the object representing the instance of the class
	   */
	  public synchronized void addToAgentList(Object agent_name,Object[] instance) {
		_agent_list.add( (String)agent_name );
		_agents.put( agent_name, instance );
	  }
  
	  /** REmove and agent from the agent list
	   * @param agent_name an object representing the name of the agent
	   */
	  public synchronized void removeFromAgentLists(Object agent_name) {
		_agents.remove( agent_name );
		_agent_list.remove( (String)agent_name );
	  }
  
		/** REmove and agent from the agent list
	   * @param agent_name an object representing the name of the agent
	   */
		public synchronized void removeFromHostsLists(Object agent_name) {
			_hosts.remove( agent_name );
		}
  
	  /** Get the object representing an instance of the an agent
	   * @param agent_name name of the agent
	   * @return an object array with the agent class in 0 and agent instance in 1
	   */
	  public synchronized Object[] getAgentInstance(Object agent_name) {
		return (Object[])_agents.get( agent_name );
	  }
  
  
	  /** Get the list of instantiated agents
	   * @return the list of agents instantiated
	   */
	  public synchronized Vector getAgentList() {
		return new Vector( _agent_list );
	  }
  
	   /** Removes the Agent from the Vector of Agents
	   * @param agent The agent to be removed
	   */
	  public void removeAgent(int agent) {
		_cell_data.remove( agent );
	  }
  
	  /** creates the table data from the profile
	 * @param file_name the profile file name
	 * @param location where the file can be found
	 */
	  public void initCellData( String location, String file_name )
	  {
		//make sure htere is nothing in the cell data table
		_cell_data.clear();
		
		if ( ! file_name.endsWith( PROFILE ) )
		{
			 file_name += PROFILE;    
		}
    
		_profile_file_name = file_name;
    
		try
		{ 
			if( !location.endsWith( java.io.File.separator ) )
			{
				location = location + java.io.File.separator;
			}
			_doc_content = fipaos.agent.profile.Profile.readProfile( location + file_name );
		}
		catch( fipaos.parser.ParserException pe )
		{
			pe.printStackTrace();
		}
		catch( java.io.FileNotFoundException fnfe )
		{
			 fnfe.printStackTrace();
		}
		//get the agents bag object from the profile
		Content content = (Content)_doc_content.getAttribute( AGENTS_BAG );
		Enumeration agent_enum = extractFromBag( content );
		while( agent_enum.hasMoreElements() )
		{
			// retrieve Content object containing agent information
			Content next_content = (Content) agent_enum.nextElement();
			Iterator it =  next_content.getAttributeNames().iterator();
			Vector attributes = new Vector();
			while( it.hasNext() )
			{
				//get the attributes from the content object
				attributes.add( it.next() );
			}//while( it.hasNext() )
			String[] attribute = new String[ attributes.size() ];
			Enumeration e = attributes.elements();
			int n = 0;
			while( e.hasMoreElements() )
			{
				attribute[n] = (String)e.nextElement();
				n++;
			}//while( e.hasMoreElements() )
			//assign them to the appropriate variable
			// we have to do this in a for loop because it is an un-ordered bag
			String class_name = null;
			String platform_profile_location = "";
			String agent_name = null;
			String owner = null;
			String start = null;
			String host = null;
			String element = null;
			Object value = null;
			Iterator iter = next_content.getAttributeNames().iterator();
			String str = null;
			//for( int count = 0; count < attribute.length; count++ )
			while( iter.hasNext() )
			{
				str = (String)iter.next();
				 value = next_content.getAttribute( str );
				 
				if( value instanceof String )
				{
					
					if ( str.equals( CLASS_NAME ) )
					{
						class_name = (String)value; 
					}
					else if ( str.equals( PLATFORM_PROFILE_LOCATION ) )
					{
						platform_profile_location = (String)value;               
					}
					else if ( str.equals( AGENT_NAME ) )
					{
						agent_name = (String)value;               
					}
					else if ( str.equals( OWNER ) )
				 	{
						owner = (String)value;                
				 	}
					else if ( str.equals( START ) )
				 	{
						start = (String)value;                
					}
					else if ( str.equals( HOST ) )
				 	{
						host = (String)value;
				 	}
				}//if( value instanceof String )
			}//while( iter.hasNext() )
			Vector columns = new Vector();
			columns.add( class_name );
			if( platform_profile_location == null || platform_profile_location.equals("") )
			{
				columns.add( _platform_profile );
			}
			else
			{
				columns.add( platform_profile_location );
			}
            //column 2
			columns.add( agent_name );
            //column 3
			columns.add( owner );
            //column 4
			columns.add( start );
            //column 5
			columns.add( host );
            
			_cell_data.add( columns );
		} // while(agent_enum.hasMoreElements())
	}
      
	  /** Shut down the agent represented by the Object
	   * @param agent The agent to be shutdown
	   */
	  public void shutdownAgent(Object agent) {
		//check if its a remote agent
		if( _hosts.containsKey( agent ) )
		{
			Object[] obj = (Object[])_hosts.get( agent );
			VMCommsInterface vm = (VMCommsInterface)obj[1];
			try
			{
				vm.shutdownagent( (String)agent );
			}
			catch(RemoteException e)
			{
				e.printStackTrace();
			}
			removeFromHostsLists( agent );
		}
		else
		{
			Object[] class_instance = null;
			class_instance = getAgentInstance( agent );
			FIPAOSAgent fipa_agent = (FIPAOSAgent)class_instance[1];
			fipa_agent.shutdown();
		}
		removeFromAgentLists( agent );
	  }
  
	  /** Gets the agent details from the loader profile and constructs a Vector of Vectors ready to populate a table to display the info
	   * @return Vector of Vectors representing the cells of a table
	   */
	  public Vector getCellData() 
	  {
	
		return _cell_data;
	  }
	  /** Starts a remote agent
	 * @param class_name Name of the class to be instanitated
	 * @param parameters parameters to be passed to the constructor( at present 3 Strings)
	 * @param host The host where the agent is to run
	 */
	  public void startRemoteAgent(final String class_name, final Object[] parameters, final String host)
	  {
		// Set default platform profile if necessary
		if( (parameters[0]) == null || ((String)parameters[0]).length() == 0 || ((String)parameters[0]) .trim().length() == 0 )
		{
			parameters[0] = _platform_profile;
		}
		 
		//get the host name without the port number
		String host_name = host.substring( 0, host.indexOf( ":" ) );
		String vm_name = VM_REMOTE + host_name;
		DIAGNOSTICS.println( "trying to start Agent on remote host " + host, this, DIAGNOSTICS.LEVEL_4);
		
		try
		{
			//construct a string which represents the remote vm stub so that it can be retrieved
			String remote_vm_name = "//" + host + "/" + vm_name;
			VMCommsInterface rvm = ( VMCommsInterface)Naming.lookup( remote_vm_name );
			DIAGNOSTICS.println( "Got a remote reference to the newly-Activatable object", this, DIAGNOSTICS.LEVEL_4);
            
			try
			{
				//create the agent on remote platform
				rvm.createAgent( class_name, parameters );
				//put the hostname into an array with a reference to the to the vm into an array
				//this is used by the gui to construct a host node later
				Object[] arr = { host_name, rvm };
				//put a reference to the VM in a hashtable
				//with the name of the agent and the host that it is running on 
				//so it can be controled at any time
				_hosts.put( parameters[1], arr );
                
			}
			catch( java.rmi.RemoteException re)
			{
				DIAGNOSTICS.println( "RemoteException occured when trying to start Agent: " + parameters[1], this, DIAGNOSTICS.LEVEL_4);
				re.printStackTrace();
			}
		} 
		catch (RemoteException e) {
			System.out.println("RemoteException: ");
			e.printStackTrace();
		}
		catch (NotBoundException nbe) {
			System.out.println("NotBoundException: ");
			nbe.printStackTrace();
		}
		catch( MalformedURLException me )
		{
			me.printStackTrace();
		}
	  }

	/** Get the hashtable of remote agents and the host vms
	 * @return the list of agents and hosts
	 */
	public Hashtable getHosts()
	{
		return _hosts;
	}
	 
	/** Starts a lovcsl agent
	   * @param class_name Name of the class to be instanitated
	   * @param parameters parameters to be passed to the constructor( at present 3 Strings)
	   * 
	   */
	  public void startLocalAgent(final String class_name,final Object[] parameters)
	  {
		// Set default platform profile if necessary
		String profile = (String) parameters[0];
		if( profile == null || profile.length() == 0 || profile.trim().length() == 0 )
		{
			parameters[0] = _platform_profile;
		}
		// "started_agent" indicator
		final Object started_agent[] = { new Boolean( false ) };
		// A thread group so that the threads belonging to an agent can be kept track of
		ThreadGroup new_tg = new ThreadGroup( (String)parameters[1] );
		_agent_threads.add( new_tg );
		//inner class to create a thread for the agent to be started in
		Runnable r = new Runnable() 
		{
		  public void run()
		  {
			Object agent = null;
			try
			{
			  // get the class to instantiate
			  Class cl = Class.forName( class_name );
              Class[] cl_arr = { Class.forName( "java.lang.String"), Class.forName( "java.lang.String") , Class.forName( "java.lang.String")  };
              Constructor cons = cl.getDeclaredConstructor( cl_arr );
			  Object[] class_instance = new Object[2];
			  class_instance[0] = class_name;
			  class_instance[1] = cons.newInstance( parameters );
			  addToAgentList( (String)parameters[1], class_instance );

			  // Indicate that the Agent has started OK to the parent thread
			  synchronized( started_agent )
			  {
				started_agent[0] = new Boolean( true );
				started_agent.notify();
			  }
			}
			catch(Exception cnf)
			{ 
              
			  cnf.printStackTrace();
			}
		  }
		};
    
		try
		{
		  Thread thread = new Thread( new_tg, r );
		  thread.start();
		  // Wait for the thread to finish starting the Agent
		  synchronized( started_agent )
		  {
			while( !( (Boolean) started_agent[0] ).booleanValue() )
			{
			  try
			  {
				started_agent.wait();
			  }
			  catch( InterruptedException e )
			  {
			  }
			}
		  }
		}
		catch(Exception cnf)
		{
		  cnf.printStackTrace();
		}
	  }
  
  
    	
     /** <br> When passed a content object this method will return an Enumeration of the
     * <br> Content objects it contains.
     * <br> This assumes that the content object passed to this method contains a List of
     * <br> Content objects i.e. the RDF class that was parsed to produce this content
     * <br> object was a Bag.
     * @param content content from which to extract
     * @return extracted contents
     */
	 private Enumeration extractFromBag(Content content)
	 {
		// populate with the Content objects retrieved as a result of this method.
		// Convert this into an Enumeration and return.
		Vector results = new Vector();
	 	  
	 	// Iterate through the attributes of this Content object
	 	Iterator i =  content.getAttributeNames().iterator();
	 	  		
	 	while(i.hasNext())
	 	{
			// Get the attribute name
			String attribute_name = (String) i.next();
	 	  		
			// Retrieve the object this referrs to.
			Object obj = content.getAttribute( attribute_name );
	 	  				
			// Check if we're retreiving a Content object
			// or a List
			if( obj instanceof List )
			{
				List list = (List) obj;	
				    
				// Iterate through the list extracting all the content objects it contains
				// and adding into the results vector.			
						
				Iterator list_iterator = list.iterator();
	 	  	 
				while(list_iterator.hasNext())
				{
				// get rid of first elem in list.  This is the bag.
                               
					Object next_obj = list_iterator.next();	
		 	   	  	
					if(next_obj instanceof Content)
					{
						Content temp_cont = (Content)next_obj;
						if (  !temp_cont.isEmpty()  )
						{
							results.addElement(next_obj); 	
						}
					}
					else
					{
						DIAGNOSTICS.println("Error: member of list is not a Content object.");
					}	
				} // while(list_iterator.hasNext())	  
								
			}
			else if( obj instanceof Content )
			{
				Content temp_cont = (Content)obj;
				if (  !temp_cont.isEmpty()  )
				{
					results.addElement(obj);
				}
			}
			else
			{
				DIAGNOSTICS.println("Error: Found " + obj.getClass().getName() );
			}
		 	   
		} // while(i.hasNext())
			// convert the Vector of results into an Enumeration and return
		return results.elements();
	} // extractFromBag(Content content)
	 	
	 
    
    /** Method that iterates through the _cell_data and starts all agents that are required to startup on paltform boot
     */
    
    private void startAgents() 
	{
		Vector row = null;
		int no_of_parameters = 3;
		int n;
		for( int i = 0; i < _cell_data.size(); i++ )
		{
			//get the current row
			row = (Vector)_cell_data.get(i);
            //set up some variables
			String class_name = (String)row.get(0);
			String name = (String)row.get(1);
            String host = (String)row.get(5);
            Object[] parameters = new Object[no_of_parameters];
            
			for(n = 0; n < no_of_parameters; n++ )
            {
                parameters[n] = (String) row.get( n + 1 );
            }
			//if the agent needs starting on platform boot
			if ( ((String)row.get( n + 1 )).equals( TRUE ) )
			{
				if( host.equals( PLATFORM_VM ) )
				{
					//start it locally
					startLocalAgent( class_name, parameters );
				}
				else
				{
					//start it remotely
					startRemoteAgent( class_name, parameters, host );
				}
			}
		}
    }
        
        
    
//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------	


    /**
     * @param args  */
    public static void main(String[] args) 
	{
		DIAGNOSTICS.setMemoryLevel( DIAGNOSTICS.LEVEL_2 );
			
		if( args.length < 2)
		{
			System.out.println( "Usage: fipaos.platform.manager platform-profile-location loader-profile-name" );
		}
		else
		{
			VMPlatformManager platform_manager = new VMPlatformManager( args[0], args[1] );
		}
	}
}